# -*- coding: utf-8 -*-
#This code is adaped from
# https://github.com/dfm/corner.py
# git hash 5c2cd63  on May 25
# Modifications by John Baker NASA-GSFC (2016)
# Modifications by Sylvain Marsat AEI (2017)
#Copyright (c) 2013-2016 Daniel Foreman-Mackey
#All rights reserved.
#
#Redistribution and use in source and binary forms, with or without
#modification, are permitted provided that the following conditions are met:
#
#1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#2. Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
#THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
#ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#The views and conclusions contained in the software and documentation are those
#of the authors and should not be interpreted as representing official policies,
#either expressed or implied, of the FreeBSD Project.

from __future__ import absolute_import, division, print_function
import sys
if sys.version_info[0] == 2:
    from future_builtins import map, filter


import logging
import math
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
from matplotlib.colors import LinearSegmentedColormap, colorConverter
from matplotlib.ticker import ScalarFormatter
from matplotlib.patches import Ellipse

try:
    from scipy.ndimage import gaussian_filter
except ImportError:
    gaussian_filter = None

__all__ = ["corner", "hist2d", "quantile"]


def corner(xs, bins=20, params_range=None, weights=None, cov=None, color="k",
           smooth=None, smooth1d=None,
           labels=None, label_kwargs=None,
           show_histograms=True, show_titles=False, title_fmt=".2f", title_kwargs=None,
           truths=None, add_truths=None, truth_color="#4682b4", add_truth_colors=None, cov_color=None,
           scale_hist=False, quantiles=None, show_quantiles=True, verbose=False, figaxes=None,
           max_n_ticks=5, top_ticks=False, use_math_text=False, resize_fig=1.,
           hist_kwargs=None, **hist2d_kwargs):
    """
    Make a *sick* corner plot showing the projections of a data set in a
    multi-dimensional space. kwargs are passed to hist2d() or used for
    `matplotlib` styling.

    Parameters
    ----------
    xs : array_like[nsamples, ndim]
        The samples. This should be a 1- or 2-dimensional array. For a 1-D
        array this results in a simple histogram. For a 2-D array, the zeroth
        axis is the list of samples and the next axis are the dimensions of
        the space.

    bins : int or array_like[ndim,]
        The number of bins to use in histograms, either as a fixed value for
        all dimensions, or as a list of integers for each dimension.

    weights : array_like[nsamples,]
        The weight of each sample. If `None` (default), samples are given
        equal weight.

    color : str
        A ``matplotlib`` style color for all histograms.

    smooth, smooth1d : float
       The standard deviation for Gaussian kernel passed to
       `scipy.ndimage.gaussian_filter` to smooth the 2-D and 1-D histograms
       respectively. If `None` (default), no smoothing is applied.

    labels : iterable (ndim,)
        A list of names for the dimensions. If a ``xs`` is a
        ``pandas.DataFrame``, labels will default to column names.

    label_kwargs : dict
        Any extra keyword arguments to send to the `set_xlabel` and
        `set_ylabel` methods.

    show_histograms : bool
        Displays 1-D histogram showing marginalized distributions and quantiles.

    show_titles : bool
        Displays a title above each 1-D histogram showing the 0.5 quantile
        with the upper and lower errors supplied by the quantiles argument.

    title_fmt : string
        The format string for the quantiles given in titles. If you explicitly
        set ``show_titles=True`` and ``title_fmt=None``, the labels will be
        shown as the titles. (default: ``.2f``)

    title_kwargs : dict
        Any extra keyword arguments to send to the `set_title` command.

    params_range : iterable (ndim,)
        A list where each element is either a length 2 tuple containing
        lower and upper bounds or a float in range (0., 1.)
        giving the fraction of samples to include in bounds, e.g.,
        [(0.,10.), (1.,5), 0.999, etc.].
        If a fraction, the bounds are chosen to be equal-tailed.

    truths : iterable (ndim,)
        A list of reference values to indicate on the plots.  Individual
        values can be omitted by using ``None``.

    truth_color : str
        A ``matplotlib`` style color for the ``truths`` makers.

    scale_hist : bool
        Should the 1-D histograms be scaled in such a way that the zero line
        is visible?

    quantiles : iterable
        A list of fractional quantiles to show on the 1-D histograms as
        vertical dashed lines.

    verbose : bool
        If true, print the values of the computed quantiles.

    plot_contours : bool
        Draw contours for dense regions of the plot.

    use_math_text : bool
        If true, then axis tick labels for very large or small exponents will
        be displayed as powers of 10 rather than using `e`.

    max_n_ticks: int
        Maximum number of ticks to try to use

    top_ticks : bool
        If true, label the top ticks of each axis

    fig : matplotlib.Figure
        Overplot onto the provided figure object.

    hist_kwargs : dict
        Any extra keyword arguments to send to the 1-D histogram plots.

    **hist2d_kwargs
        Any remaining keyword arguments are sent to `corner.hist2d` to generate
        the 2-D histogram plots.

    """
    if quantiles is None:
        quantiles = []
    if title_kwargs is None:
        title_kwargs = dict()
    if label_kwargs is None:
        label_kwargs = dict()

    # Try filling in labels from pandas.DataFrame columns.
    if labels is None:
        try:
            labels = xs.columns
        except AttributeError:
            pass

    # Deal with 1D sample lists.
    xs = np.atleast_1d(xs)
    if len(xs.shape) == 1:
        xs = np.atleast_2d(xs)
    else:
        assert len(xs.shape) == 2, "The input sample array must be 1- or 2-D."
        xs = xs.T
    assert xs.shape[0] <= xs.shape[1], "I don't believe that you want more " \
                                       "dimensions than samples!"

    # Parse the weight array.
    if weights is not None:
        weights = np.asarray(weights)
        if weights.ndim != 1:
            raise ValueError("Weights must be 1-D")
        if xs.shape[1] != weights.shape[0]:
            raise ValueError("Lengths of weights must match number of samples")

    # Parse the parameter ranges.
    if params_range is None:
        if "extents" in hist2d_kwargs:
            logging.warn("Deprecated keyword argument 'extents'. "
                         "Use 'params_range' instead.")
            params_range = hist2d_kwargs.pop("extents")
        else:
            params_range = [[x.min(), x.max()] for x in xs]
            # Check for parameters that never change.
            m = np.array([e[0] == e[1] for e in params_range], dtype=bool)
            if np.any(m):
                raise ValueError(("It looks like the parameter(s) in "
                                  "column(s) {0} have no dynamic range. "
                                  "Please provide a `params_range` argument.")
                                 .format(", ".join(map(
                                     "{0}".format, np.arange(len(m))[m]))))

    else:
        # If any of the extents are percentiles, convert them to ranges.
        # Also make sure it's a normal list.
        params_range = list(params_range)
        for i, pr in enumerate(params_range):
            if pr is None:
                params_range[i] = [xs[i].min(), xs[i].max()]
            else:
                try:
                    emin, emax = params_range[i]
                except TypeError:
                    q = [0.5 - 0.5*params_range[i], 0.5 + 0.5*params_range[i]]
                    params_range[i] = quantile(xs[i], q, weights=weights)

    if len(params_range) != xs.shape[0]:
        raise ValueError("Dimension mismatch between samples and params_range")

    # Parse the bin specifications.
    try:
        bins = [int(bins) for _ in params_range]
    except TypeError:
        if len(bins) != len(params_range):
            raise ValueError("Dimension mismatch between bins and params_range")

    # Some magic numbers for pretty axis layout.
    K = len(xs)
    if show_histograms:    # total number of plots: grid nplot*nplot
        nplot = K
    else:
        nplot = K-1
    factor = 2.0           # size of one side of one panel
    lbdim = 0.5 * factor   # size of left/bottom margin
    trdim = 0.2 * factor   # size of top/right margin
    whspace = 0.05         # w/hspace size
    plotdim = factor * nplot + factor * (nplot - 1.) * whspace
    dim = lbdim + plotdim + trdim

    # Create a new figure if one wasn't provided.
    # if fig is None:
    #     fig, axes = plt.subplots(nplot, nplot, figsize=(dim, dim))
    # else:
    #     print(type(fig.axes))
    #     try:
    #         axes = np.array(fig.axes).reshape((nplot, nplot))
    #     except:
    #         raise ValueError("Provided figure has {0} axes, but data has "
    #                          "dimensions K={1}".format(len(fig.axes), nplot))
    if figaxes is None:
        fig, axes = plt.subplots(nplot, nplot, figsize=(resize_fig*dim, resize_fig*dim))
    else:
        fig, axes = figaxes


    #idea is to pass in covariance, otherwise concoct something from the 1-sigma range.
    # if(cov==None):
    #     print("concocting covar elements from 1-sigma ranges")
    #     cov=np.zeros((K,K))
    #     for k in np.arange(K):
    #         q_16, q_50, q_84 = quantile(xs[k], [0.16, 0.5, 0.84],weights=weights)
    #         deltax=(q_84-q_16)/2.0
    #         cov[k,k]=deltax**2
    # print("cov=",cov)

    # Format the figure.
    lb = lbdim / dim
    tr = (lbdim + plotdim) / dim
    fig.subplots_adjust(left=lb, bottom=lb, right=tr, top=tr,
                        wspace=whspace, hspace=whspace)

    # If asked to show the 1D histograms
    if show_histograms:
        # Set up the default histogram keywords.
        if hist_kwargs is None:
            hist_kwargs = dict()
        hist_kwargs["color"] = hist_kwargs.get("color", color)
        if smooth1d is None:
            hist_kwargs["histtype"] = hist_kwargs.get("histtype", "step")
        for i, x in enumerate(xs):
            # Deal with masked arrays.
            if hasattr(x, "compressed"):
                x = x.compressed()

            if np.shape(xs)[0] == 1:
                ax = axes
            else:
                ax = axes[i, i]
            # Plot the histograms.
            if smooth1d is None:
                n, _, _ = ax.hist(x, bins=bins[i], weights=weights, density=True,
                                  range=np.sort(params_range[i]), **hist_kwargs)
            else:
                if gaussian_filter is None:
                    raise ImportError("Please install scipy for smoothing")
                n, b = np.histogram(x, bins=bins[i], weights=weights, density=True,
                                    range=np.sort(params_range[i]))
                n = gaussian_filter(n, smooth1d)
                x0 = np.array(list(zip(b[:-1], b[1:]))).flatten()
                y0 = np.array(list(zip(n, n))).flatten()
                ax.plot(x0, y0, **hist_kwargs)

            if truths is not None and truths[i] is not None:
                ax.axvline(truths[i], color=truth_color)

            if add_truths is not None:
                for add_truth, add_truth_color in zip(add_truths, add_truth_colors):
                     if add_truth[i] is not None:
                         ax.axvline(add_truth[i], color=add_truth_color)

            # Plot quantiles if wanted.
            if show_quantiles:
                if len(quantiles) > 0:
                    qvalues = quantile(x, quantiles, weights=weights)
                    for q in qvalues:
                        ax.axvline(q, ls="dashed", color=color)

                    if verbose:
                        print("Quantiles:")
                        for item in zip(quantiles, qvalues):
                            print(item)

            if show_titles:
                title = None
                if title_fmt is not None:
                    # Compute the quantiles for the title. This might redo
                    # unneeded computation but who cares.
                    q_16, q_50, q_84 = quantile(x, [0.16, 0.5, 0.84],
                                                weights=weights)
                    q_m, q_p = q_50-q_16, q_84-q_50

                    # Format the quantile display.
                    fmt = "{{0:{0}}}".format(title_fmt).format
                    title = r"${{{0}}}_{{-{1}}}^{{+{2}}}$"
                    title = title.format(fmt(q_50), fmt(q_m), fmt(q_p))

                    # Add in the column name if it's given.
                    if labels is not None:
                        title = "{0} = {1}".format(labels[i], title)

                elif labels is not None:
                    title = "{0}".format(labels[i])

                if title is not None:
                    ax.set_title(title, **title_kwargs)

            # Set up the axes.
            ax.set_xlim(params_range[i])
            if scale_hist:
                maxn = np.max(n)
                ax.set_ylim(-0.1 * maxn, 1.1 * maxn)
            else:
                ax.set_ylim(0, 1.1 * np.max(n))
            ax.set_yticklabels([])
            ax.xaxis.set_major_locator(MaxNLocator(max_n_ticks, prune="lower"))

            if i < K - 1:
                if top_ticks:
                    ax.xaxis.set_ticks_position("top")
                    [l.set_rotation(45) for l in ax.get_xticklabels()]
                else:
                    ax.set_xticklabels([])
            else:
                [l.set_rotation(45) for l in ax.get_xticklabels()]
                if labels is not None:
                    ax.set_xlabel(labels[i], **label_kwargs)
                    ax.xaxis.set_label_coords(0.5, -0.3)

                # use MathText for axes ticks
                ax.xaxis.set_major_formatter(
                    ScalarFormatter(useMathText=use_math_text))

            for j, y in enumerate(xs):
                if np.shape(xs)[0] == 1:
                    ax = axes
                else:
                    ax = axes[i, j]
                if j > i:
                    ax.set_frame_on(False)
                    ax.set_xticks([])
                    ax.set_yticks([])
                    continue
                elif j == i:
                    continue

                # Deal with masked arrays.
                if hasattr(y, "compressed"):
                    y = y.compressed()

                hist2d(y, x, ax=ax, range=[params_range[j], params_range[i]], weights=weights,
                       color=color, smooth=smooth, bins=[bins[j], bins[i]],
                       **hist2d_kwargs)

                # Add an error ellipse based on the provided covariance matrix
                if cov is not None:
                    # #center
                    # cx=truths[j]#need to add checking for availability of truths?
                    # cy=truths[i]
                    # #ang=math.acos(cov[0,1]/math.sqrt(cov[0,0]*cov[1,1]))*180/math.pi
                    # print (j,i,labels[j],labels[i],"center=",cx,cy)
                    # N_thetas=60
                    # dtheta=2.0*math.pi/(N_thetas-1)
                    # thetas=np.arange(0,(2.0*math.pi+dtheta),dtheta)
                    # Cplus=(cov[i,i]+cov[j,j])/2.0
                    # Cminus=(-cov[i,i]+cov[j,j])/2.0
                    # print("cov[ii],cov[ij],cov[jj],Cplus,Cminus:",cov[i,i],cov[i,j],cov[j,j],Cplus,Cminus)
                    # ang=-math.pi/4.
                    # root=cov[i,j]/math.sqrt(cov[i,i]*cov[j,j])
                    # acoeff=math.sqrt(1-root)
                    # bcoeff=math.sqrt(1+root)
                    # xcoeff=math.sqrt(cov[j,j])
                    # ycoeff=math.sqrt(cov[i,i])
                    # print("a2,b2",acoeff*acoeff,bcoeff*bcoeff)
                    # print("a,b,ang, xcoeff,ycoeff, root=",acoeff,bcoeff,ang,xcoeff,ycoeff,root)
                    # elxs=[cx+xcoeff*(acoeff*math.cos(th)*math.cos(ang)-bcoeff*math.sin(th)*math.sin(ang)) for th in thetas]
                    # elys=[cy+ycoeff*(acoeff*math.cos(th)*math.sin(ang)+bcoeff*math.sin(th)*math.cos(ang)) for th in thetas]
                    # #print (thetas)
                    # #print (elxs)
                    # #print (elys)
                    # ax.plot(elxs,elys,color='r')
                    #
                    #center
                    if truths is None:
                        raise ValueError('Error: truths is None but is required to plot Fisher ellipses.')
                    cx = truths[j] # need to add checking for availability of truths?
                    cy = truths[i]
                    N_thetas = 60
                    dtheta = 2.0*math.pi/(N_thetas-1)
                    thetas = np.arange(0,(2.0*math.pi+dtheta),dtheta)
                    ang = -math.pi/4.
                    root = cov[i,j]/math.sqrt(cov[i,i]*cov[j,j])
                    if(root>1): root = 1
                    if(root<-1): root = -1
                    acoeff = math.sqrt(1-root)
                    bcoeff = math.sqrt(1+root)
                    xcoeff = math.sqrt(cov[j,j])
                    ycoeff = math.sqrt(cov[i,i])
                    if "levels" in hist2d_kwargs:
                        levels= hist2d_kwargs["levels"]
                    else:
                        levels== 1.0 - np.exp(-0.5 * np.arange(0.5, 2.1, 0.5) ** 2)
                    for xlev in levels:
                        # in the next line we convert the credibility limit
                        # to a "sigma" limit for a 2-d normal
                        # this becomes a scale-factor for the error ellipse
                        # 1-exp(x^2/(-2)=y
                        # -2*log(1-y)=x^2
                        lev_fac = math.sqrt( -2 * math.log( 1 - xlev ) )
                        elxs = [cx+lev_fac*xcoeff*(acoeff*math.cos(th)*math.cos(ang)-bcoeff*math.sin(th)*math.sin(ang)) for th in thetas]
                        elys = [cy+lev_fac*ycoeff*(acoeff*math.cos(th)*math.sin(ang)+bcoeff*math.sin(th)*math.cos(ang)) for th in thetas]
                        if cov_color is None:
                            cov_color='k'
                        ax.plot(elxs, elys, color=cov_color)

                if truths is not None:
                    if truths[i] is not None and truths[j] is not None:
                        ax.plot(truths[j], truths[i], "s", color=truth_color)
                    if truths[j] is not None:
                        ax.axvline(truths[j], color=truth_color)
                    if truths[i] is not None:
                        ax.axhline(truths[i], color=truth_color)

                if add_truths is not None:
                    for add_truth, add_truth_color in zip(add_truths, add_truth_colors):
                        if add_truth[i] is not None and add_truth[j] is not None:
                            ax.plot(add_truth[j], add_truth[i], "s", color=add_truth_color)
                        if add_truth[j] is not None:
                            ax.axvline(add_truth[j], color=add_truth_color)
                        if add_truth[i] is not None:
                            ax.axhline(add_truth[i], color=add_truth_color)

                ax.xaxis.set_major_locator(MaxNLocator(max_n_ticks, prune="lower"))
                ax.yaxis.set_major_locator(MaxNLocator(max_n_ticks, prune="lower"))

                if i < K - 1:
                    ax.set_xticklabels([])
                else:
                    [l.set_rotation(45) for l in ax.get_xticklabels()]
                    if labels is not None:
                        ax.set_xlabel(labels[j], **label_kwargs)
                        ax.xaxis.set_label_coords(0.5, -0.3)

                    # use MathText for axes ticks
                    ax.xaxis.set_major_formatter(
                        ScalarFormatter(useMathText=use_math_text))

                if j > 0:
                    ax.set_yticklabels([])
                else:
                    [l.set_rotation(45) for l in ax.get_yticklabels()]
                    if labels is not None:
                        ax.set_ylabel(labels[i], **label_kwargs)
                        ax.yaxis.set_label_coords(-0.3, 0.5)

                    # use MathText for axes ticks
                    ax.yaxis.set_major_formatter(
                        ScalarFormatter(useMathText=use_math_text))

    # If not asking for the histograms to be plotted
    else:
        for i, x in enumerate(xs):
            if i==0:
                continue
            # Deal with masked arrays.
            if hasattr(x, "compressed"):
                x = x.compressed()

            for j, y in enumerate(xs):
                if j==K-1:
                    continue
                if np.shape(xs)[0] == 2:
                    ax = axes
                else:
                    ax = axes[i-1, j]
                if j > i-1 :
                    ax.set_frame_on(False)
                    ax.set_xticks([])
                    ax.set_yticks([])
                    continue
                elif j == i:
                    continue

                # Deal with masked arrays.
                if hasattr(y, "compressed"):
                    y = y.compressed()

                hist2d(y, x, ax=ax, range=[params_range[j], params_range[i]], weights=weights,
                       color=color, smooth=smooth, bins=[bins[j], bins[i]],
                       **hist2d_kwargs)

                # Add an error ellipse based on the provided covariance matrix
                if cov is not None:
                    # #center
                    # cx=truths[j]#need to add checking for availability of truths?
                    # cy=truths[i]
                    # #ang=math.acos(cov[0,1]/math.sqrt(cov[0,0]*cov[1,1]))*180/math.pi
                    # print (j,i,labels[j],labels[i],"center=",cx,cy)
                    # N_thetas=60
                    # dtheta=2.0*math.pi/(N_thetas-1)
                    # thetas=np.arange(0,(2.0*math.pi+dtheta),dtheta)
                    # Cplus=(cov[i,i]+cov[j,j])/2.0
                    # Cminus=(-cov[i,i]+cov[j,j])/2.0
                    # print("cov[ii],cov[ij],cov[jj],Cplus,Cminus:",cov[i,i],cov[i,j],cov[j,j],Cplus,Cminus)
                    # ang=-math.pi/4.
                    # root=cov[i,j]/math.sqrt(cov[i,i]*cov[j,j])
                    # acoeff=math.sqrt(1-root)
                    # bcoeff=math.sqrt(1+root)
                    # xcoeff=math.sqrt(cov[j,j])
                    # ycoeff=math.sqrt(cov[i,i])
                    # print("a2,b2",acoeff*acoeff,bcoeff*bcoeff)
                    # print("a,b,ang, xcoeff,ycoeff, root=",acoeff,bcoeff,ang,xcoeff,ycoeff,root)
                    # elxs=[cx+xcoeff*(acoeff*math.cos(th)*math.cos(ang)-bcoeff*math.sin(th)*math.sin(ang)) for th in thetas]
                    # elys=[cy+ycoeff*(acoeff*math.cos(th)*math.sin(ang)+bcoeff*math.sin(th)*math.cos(ang)) for th in thetas]
                    # #print (thetas)
                    # #print (elxs)
                    # #print (elys)
                    # ax.plot(elxs,elys,color='r')
                    #
                    #center
                    cx = truths[j] # need to add checking for availability of truths?
                    cy = truths[i]
                    N_thetas = 60
                    dtheta = 2.0*math.pi/(N_thetas-1)
                    thetas = np.arange(0,(2.0*math.pi+dtheta),dtheta)
                    ang = -math.pi/4.
                    root = cov[i,j]/math.sqrt(cov[i,i]*cov[j,j])
                    if(root>1): root = 1
                    if(root<-1): root = -1
                    acoeff = math.sqrt(1-root)
                    bcoeff = math.sqrt(1+root)
                    xcoeff = math.sqrt(cov[j,j])
                    ycoeff = math.sqrt(cov[i,i])
                    if "levels" in hist2d_kwargs:
                        levels= hist2d_kwargs["levels"]
                    else:
                        levels== 1.0 - np.exp(-0.5 * np.arange(0.5, 2.1, 0.5) ** 2)
                    for xlev in levels:
                        # in the next line we convert the credibility limit
                        # to a "sigma" limit for a 2-d normal
                        # this becomes a scale-factor for the error ellipse
                        # 1-exp(x^2/(-2)=y
                        # -2*log(1-y)=x^2
                        lev_fac = math.sqrt( -2 * math.log( 1 - xlev ) )
                        elxs = [cx+lev_fac*xcoeff*(acoeff*math.cos(th)*math.cos(ang)-bcoeff*math.sin(th)*math.sin(ang)) for th in thetas]
                        elys = [cy+lev_fac*ycoeff*(acoeff*math.cos(th)*math.sin(ang)+bcoeff*math.sin(th)*math.cos(ang)) for th in thetas]
                        if cov_color is None:
                            cov_color='k'
                        ax.plot(elxs, elys, color=cov_color)

                if truths is not None:
                    if truths[i] is not None and truths[j] is not None:
                        ax.plot(truths[j], truths[i], "s", color=truth_color)
                    if truths[j] is not None:
                        ax.axvline(truths[j], color=truth_color)
                    if truths[i] is not None:
                        ax.axhline(truths[i], color=truth_color)

                if add_truths is not None:
                    for add_truth, add_truth_color in zip(add_truths, add_truth_colors):
                        if add_truth[i] is not None and add_truth[j] is not None:
                            ax.plot(add_truth[j], add_truth[i], "s", color=add_truth_color)
                        if add_truth[j] is not None:
                            ax.axvline(add_truth[j], color=add_truth_color)
                        if add_truth[i] is not None:
                            ax.axhline(add_truth[i], color=add_truth_color)

                ax.xaxis.set_major_locator(MaxNLocator(max_n_ticks, prune="lower"))
                ax.yaxis.set_major_locator(MaxNLocator(max_n_ticks, prune="lower"))

                if i < K - 1:
                    ax.set_xticklabels([])
                else:
                    [l.set_rotation(45) for l in ax.get_xticklabels()]
                    if labels is not None:
                        ax.set_xlabel(labels[j], **label_kwargs)
                        ax.xaxis.set_label_coords(0.5, -0.3)

                    # use MathText for axes ticks
                    ax.xaxis.set_major_formatter(
                        ScalarFormatter(useMathText=use_math_text))

                if j > 0:
                    ax.set_yticklabels([])
                else:
                    [l.set_rotation(45) for l in ax.get_yticklabels()]
                    if labels is not None:
                        ax.set_ylabel(labels[i], **label_kwargs)
                        ax.yaxis.set_label_coords(-0.3, 0.5)

                    # use MathText for axes ticks
                    ax.yaxis.set_major_formatter(
                        ScalarFormatter(useMathText=use_math_text))

    return fig, axes


def quantile(x, q, weights=None):
    """
    Compute sample quantiles with support for weighted samples.

    Note
    ----
    When ``weights`` is ``None``, this method simply calls numpy's percentile
    function with the values of ``q`` multiplied by 100.

    Parameters
    ----------
    x : array_like[nsamples,]
       The samples.

    q : array_like[nquantiles,]
       The list of quantiles to compute. These should all be in the range
       ``[0, 1]``.

    weights : Optional[array_like[nsamples,]]
        An optional weight corresponding to each sample. These

    Returns
    -------
    quantiles : array_like[nquantiles,]
        The sample quantiles computed at ``q``.

    Raises
    ------
    ValueError
        For invalid quantiles; ``q`` not in ``[0, 1]`` or dimension mismatch
        between ``x`` and ``weights``.

    """
    x = np.atleast_1d(x)
    q = np.atleast_1d(q)

    if np.any(q < 0.0) or np.any(q > 1.0):
        raise ValueError("Quantiles must be between 0 and 1")

    if weights is None:
        return np.percentile(x, 100.0 * q)
    else:
        weights = np.atleast_1d(weights)
        if len(x) != len(weights):
            raise ValueError("Dimension mismatch: len(weights) != len(x)")
        idx = np.argsort(x)
        sw = weights[idx]
        cdf = np.cumsum(sw)[:-1]
        cdf /= cdf[-1]
        cdf = np.append(0, cdf)
        return np.interp(q, cdf, x[idx]).tolist()


def hist2d(x, y, bins=20, range=None, weights=None, levels=None, smooth=None,
           ax=None, color=None, plot_datapoints=True, plot_density=True,
           plot_contours=True, no_fill_contours=False, fill_contours=False,
           contour_kwargs=None, contourf_kwargs=None, data_kwargs=None,
           **kwargs):
    """
    Plot a 2-D histogram of samples.

    Parameters
    ----------
    x : array_like[nsamples,]
       The samples.

    y : array_like[nsamples,]
       The samples.

    levels : array_like
        The contour levels to draw.

    ax : matplotlib.Axes
        A axes instance on which to add the 2-D histogram.

    plot_datapoints : bool
        Draw the individual data points.

    plot_density : bool
        Draw the density colormap.

    plot_contours : bool
        Draw the contours.

    no_fill_contours : bool
        Add no filling at all to the contours (unlike setting
        ``fill_contours=False``, which still adds a white fill at the densest
        points).

    fill_contours : bool
        Fill the contours.

    contour_kwargs : dict
        Any additional keyword arguments to pass to the `contour` method.

    contourf_kwargs : dict
        Any additional keyword arguments to pass to the `contourf` method.

    data_kwargs : dict
        Any additional keyword arguments to pass to the `plot` method when
        adding the individual data points.

    """
    if ax is None:
        ax = plt.gca()

    # Set the default range based on the data range if not provided.
    if range is None:
        if "extent" in kwargs:
            logging.warn("Deprecated keyword argument 'extent'. "
                         "Use 'range' instead.")
            range = kwargs["extent"]
        else:
            range = [[x.min(), x.max()], [y.min(), y.max()]]

    # Set up the default plotting arguments.
    if color is None:
        color = "k"

    # Choose the default "sigma" contour levels.
    if levels is None:
        levels = 1.0 - np.exp(-0.5 * np.arange(0.5, 2.1, 0.5) ** 2)

    # This is the color map for the density plot, over-plotted to indicate the
    # density of the points near the center.
    density_cmap = LinearSegmentedColormap.from_list(
        "density_cmap", [color, (1, 1, 1, 0)])

    # This color map is used to hide the points at the high density areas.
    white_cmap = LinearSegmentedColormap.from_list(
        "white_cmap", [(1, 1, 1), (1, 1, 1)], N=2)

    # This "color map" is the list of colors for the contour levels if the
    # contours are filled.
    rgba_color = colorConverter.to_rgba(color)
    contour_cmap = [list(rgba_color) for l in levels] + [rgba_color]
    for i, l in enumerate(levels):
        contour_cmap[i][-1] *= float(i) / (len(levels)+1)

    # We'll make the 2D histogram to directly estimate the density.
    try:
        H, X, Y = np.histogram2d(x.flatten(), y.flatten(), bins=bins,
                                 range=list(map(np.sort, range)),
                                 weights=weights)
    except ValueError:
        raise ValueError("It looks like at least one of your sample columns "
                         "have no dynamic range. You could try using the "
                         "'range' argument.")

    if smooth is not None:
        if gaussian_filter is None:
            raise ImportError("Please install scipy for smoothing")
        H = gaussian_filter(H, smooth)

    # Compute the density levels.
    Hflat = H.flatten()
    inds = np.argsort(Hflat)[::-1]
    Hflat = Hflat[inds]
    sm = np.cumsum(Hflat)
    sm /= sm[-1]
    V = np.empty(len(levels))
    for i, v0 in enumerate(levels):
        try:
            V[i] = Hflat[sm <= v0][-1]
        except:
            V[i] = Hflat[0]
    V.sort()
    m = np.diff(V) == 0
    if np.any(m):
        logging.warning("Too few points to create valid contours")
    while np.any(m):
        V[np.where(m)[0][0]] *= 1.0 - 1e-4
        m = np.diff(V) == 0
    V.sort()

    # Compute the bin centers.
    X1, Y1 = 0.5 * (X[1:] + X[:-1]), 0.5 * (Y[1:] + Y[:-1])

    # Extend the array for the sake of the contours at the plot edges.
    H2 = H.min() + np.zeros((H.shape[0] + 4, H.shape[1] + 4))
    H2[2:-2, 2:-2] = H
    H2[2:-2, 1] = H[:, 0]
    H2[2:-2, -2] = H[:, -1]
    H2[1, 2:-2] = H[0]
    H2[-2, 2:-2] = H[-1]
    H2[1, 1] = H[0, 0]
    H2[1, -2] = H[0, -1]
    H2[-2, 1] = H[-1, 0]
    H2[-2, -2] = H[-1, -1]
    X2 = np.concatenate([
        X1[0] + np.array([-2, -1]) * np.diff(X1[:2]),
        X1,
        X1[-1] + np.array([1, 2]) * np.diff(X1[-2:]),
    ])
    Y2 = np.concatenate([
        Y1[0] + np.array([-2, -1]) * np.diff(Y1[:2]),
        Y1,
        Y1[-1] + np.array([1, 2]) * np.diff(Y1[-2:]),
    ])

    if plot_datapoints:
        if data_kwargs is None:
            data_kwargs = dict()
        data_kwargs["color"] = data_kwargs.get("color", color)
        data_kwargs["ms"] = data_kwargs.get("ms", 2.0)
        data_kwargs["mec"] = data_kwargs.get("mec", "none")
        data_kwargs["alpha"] = data_kwargs.get("alpha", 0.1)
        ax.plot(x, y, "o", zorder=-1, rasterized=True, **data_kwargs)

    # Plot the base fill to hide the densest data points.
    if (plot_contours or plot_density) and not no_fill_contours:
        ax.contourf(X2, Y2, H2.T, [V.min(), H.max()],
                    cmap=white_cmap, antialiased=False)

    if plot_contours and fill_contours:
        if contourf_kwargs is None:
            contourf_kwargs = dict()
        contourf_kwargs["colors"] = contourf_kwargs.get("colors", contour_cmap)
        contourf_kwargs["antialiased"] = contourf_kwargs.get("antialiased",
                                                             False)
        ax.contourf(X2, Y2, H2.T, np.concatenate([[0], V, [H.max()*(1+1e-4)]]),
                    **contourf_kwargs)

    # Plot the density map. This can't be plotted at the same time as the
    # contour fills.
    elif plot_density:
        ax.pcolor(X, Y, H.max() - H.T, cmap=density_cmap)

    # Plot the contour edge colors.
    if plot_contours:
        if contour_kwargs is None:
            contour_kwargs = dict()
        contour_kwargs["colors"] = contour_kwargs.get("colors", color)
        ax.contour(X2, Y2, H2.T, V, **contour_kwargs)

    ax.set_xlim(range[0])
    ax.set_ylim(range[1])



#
# Copyright (C) 2019 Sylvain Marsat.
#
#


"""
    Cython definitions for generic waveform tools.
"""

import numpy as np
cimport numpy as np

from lisabeta.pyconstants cimport *
from lisabeta.struct.pystruct cimport *

cdef extern from "tools.h":

    double complex SpinWeightedSphericalHarmonic (
        double theta,
        double phi,
        int s,
        int l,
        int m
    );

    int BuildWaveformGeomFrequencyGrid (
        real_vector** Mfreq,
        const double Mfmin,
        const double Mfmax,
        const double eta,
        const double acc
    );

    int BuildResponseFrequencyGrid (
      real_vector** freq,
      real_matrix* tfspline,
      const double f_min,
      const double f_max,
      const double Deltat_max,
      const double Deltaf_max,
      const int nptlogmin
    );

    int BuildFrequencyGrid (
      real_vector** freq,
      const double f_min,
      const double f_max,
      const double M,
      const double q,
      const double Deltat_max,
      const double Deltaf_max,
      const double DeltalnMf_max,
      const double acc,
      const int nptlogmin
    );

    int BuildMergedGrid (
      real_vector** grid,
      real_vector* grid1,
      real_vector* grid2,
      const int usedeltaxmin,
      const double deltaxmin,
      const int usedeltalnxmin,
      const double deltalnxmin
    );

    double LinearInterpMultiMode3ChannelsResidualNorm(
      complex_array_3d* alpha0,
      complex_array_3d* alpha1,
      complex_array_3d* w0,
      complex_array_3d* w1,
      complex_array_3d* w2
    );



#
# Copyright (C) 2019 Sylvain Marsat.
#
#


"""
    Standalone IMRPhenomD inspiral-merger-ringdown GW waveform model
    for binary black hole coalescences.
"""

import numpy as np
cimport numpy as np
from math import cos

from lisabeta.pyconstants cimport *
from lisabeta.struct.pystruct cimport *

# cdef extern from "IMRPhenomD_internals.h":
#
#     ctypedef struct AmpPhaseFDWaveform:
#         double* freq;
#         double* amp;
#         double* phase;
#         size_t length;
#
#     AmpPhaseFDWaveform* CreateAmpPhaseFDWaveform(
#         size_t length
#     );
#
#     void DestroyAmpPhaseFDWaveform(AmpPhaseFDWaveform* wf);
#
#     ctypedef struct RealVector:
#         double* data;
#         size_t length;
#
#     RealVector* CreateRealVector(
#         size_t length
#     );
#
#     void DestroyRealVector(RealVector* v);

cdef extern from "IMRPhenomD_internals.h":
    # cdef double _MSUN_SI "MSUN_SI"
    # cdef double _MTSUN_SI "MTSUN_SI"
    # cdef double _MRSUN_SI "MRSUN_SI"
    # cdef double _PC_SI "PC_SI"

    ctypedef struct COMPLEX16FrequencySeries:
        double complex *data;
        char *name;
        long epoch;
        double f0;
        double deltaF;
        size_t length;

    COMPLEX16FrequencySeries *CreateCOMPLEX16FrequencySeries(
        const char *name,
        long epoch,
        double f0,
        double deltaF,
        size_t length
    );

    COMPLEX16FrequencySeries *ResizeCOMPLEX16FrequencySeries(COMPLEX16FrequencySeries *fs, size_t length);
    void DestroyCOMPLEX16FrequencySeries(COMPLEX16FrequencySeries *fs);

cdef extern from "IMRPhenomD_internals.h":
    ctypedef struct ModGRParams:
        double alpha;
    ctypedef struct ExtraParams:
        int use_buggy_LAL_tpeak;
        # int scale_mode_freq;

# # Expose some useful constants hidden in C-space
# MSUN_SI = _MSUN_SI
# MTSUN_SI = _MTSUN_SI
# MRSUN_SI = _MRSUN_SI
# PC_SI = _PC_SI

cdef extern from "IMRPhenomD.h":
    int IMRPhenomDGenerateFD(
        COMPLEX16FrequencySeries **htilde, # [out] FD waveform
        const double phi0,                 # Orbital phase at fRef (rad)
        const double fRef_in,              # reference frequency (Hz)
        const double deltaF,               # Sampling frequency (Hz)
        const double m1_SI,                # Mass of companion 1 (kg)
        const double m2_SI,                # Mass of companion 2 (kg)
        const double chi1,                 # Aligned-spin parameter of companion 1
        const double chi2,                 # Aligned-spin parameter of companion 2
        const double f_min,                # Starting GW frequency (Hz)
        const double f_max,                # End frequency; 0 defaults to Mf = \ref f_CUT
        const double distance              # Distance of source (m)
    );

cdef extern from "IMRPhenomD.h":
    int IMRPhenomDGenerateh22FDAmpPhase(
        AmpPhaseFDWaveform** h22,           # [out] FD waveform
        real_vector** tf,                   # [out] tf from analytic derivative of the phase
        double* fpeak,                      # [out] Approximate peak frequency (Hz)
        double* tpeak,                      # [out] tf at peak frequency (s)
        double* phipeak,                    # [out] phase 22 at peak frequency
        double* fstart,                     # [out] Starting frequency (Hz)
        double* tstart,                     # [out] tf at starting frequency (s)
        double* phistart,                   # [out] phase 22 at starting frequency
        real_vector* freq,                  # Input: frequencies (Hz) on which to evaluate h22 FD - will be copied in the output AmpPhaseFDWaveform. Frequencies exceeding max freq covered by PhenomD will be given 0 amplitude and phase.
        const double m1,                    # Mass of companion 1 (solar masses)
        const double m2,                    # Mass of companion 2 (solar masses)
        const double chi1,                  # Aligned-spin parameter of companion 1
        const double chi2,                  # Aligned-spin parameter of companion 2
        const double distance,              # Distance of source (Mpc)
        const double tRef,                  # Time at fRef_for_tRef (s)
        const double phiRef,                # Orbital phase at fRef_for_phiRef (rad)
        const double fRef_for_tRef_in,      # Ref. frequency (Hz) for tRef
        const double fRef_for_phiRef_in,    # Ref. frequency (Hz) for phiRef
        const int force_phiRef_fRef,        # Flag to force phiRef at fRef after adjusting tRef
        const double Deltat,                # Time shift (s) applied a posteriori
        const ExtraParams* extraparams,     # Additional parameters
        const ModGRParams* modgrparams      # Modified GR parameters
    );

cdef extern from "IMRPhenomD.h":
    double IMRPhenomDComputeTimeOfFrequency(
        const double f,                     # Input frequency (Hz): we compute t(f)
        const double m1,                    # Mass of companion 1 (solar masses)
        const double m2,                    # Mass of companion 2 (solar masses)
        const double chi1,                  # Aligned-spin parameter of companion 1
        const double chi2,                  # Aligned-spin parameter of companion 2
        const double distance,              # Distance of source (Mpc)
        const double tRef,                  # Time at fRef_for_tRef (s)
        const double phiRef,                # Orbital phase at fRef_for_phiRef (rad)
        const double fRef_for_tRef_in,      # Ref. frequency (Hz) for tRef
        const double fRef_for_phiRef_in,    # Ref. frequency (Hz) for phiRef
        const int force_phiRef_fRef,        # Flag to force phiRef at fRef after adjusting tRef
        const double Deltat,                # Time shift (s) applied a posteriori
        const ExtraParams* extraparams,     # Additional parameters
        const ModGRParams* modgrparams      # Modified GR parameters
    );
    double IMRPhenomDComputeInverseFrequencyOfTime(
        const double t,                     # Input time (s): we compute f(t)
        const double f_estimate,            # Estimate of f(t), use to initialize
        const double t_acc,                 # Target accuracy of t(f) where to stop refining f
        const double m1,                    # Mass of companion 1 (solar masses)
        const double m2,                    # Mass of companion 2 (solar masses)
        const double chi1,                  # Aligned-spin parameter of companion 1
        const double chi2,                  # Aligned-spin parameter of companion 2
        const double distance,              # Distance of source (Mpc)
        const double tRef,                  # Time at fRef_for_tRef (s)
        const double phiRef,                # Orbital phase at fRef_for_phiRef (rad)
        const double fRef_for_tRef_in,      # Ref. frequency (Hz) for tRef
        const double fRef_for_phiRef_in,    # Ref. frequency (Hz) for phiRef
        const int force_phiRef_fRef,        # Flag to force phiRef at fRef after adjusting tRef
        const double Deltat,                # Time shift (s) applied a posteriori
        const int max_ite,                  # Maximal number of iterations in bisection
        const ExtraParams* extraparams,     # Additional parameters
        const ModGRParams* modgrparams      # Modified GR parameters
    );


cdef extern from "struct.h":
    int ERROR_EINVAL
    int ERROR_EDOM
    int ERROR_EFAULT
    int ERROR_EFUNC
    int ERROR_ENOMEM

cpdef double check_error(double value):
    if value in [ERROR_EINVAL, ERROR_EDOM, ERROR_EFAULT, ERROR_EFUNC, ERROR_ENOMEM]:
        raise ValueError
    return value

cdef class IMRPhenomD:
    """ Generate IMRPhenomD inspiral-merger-ringdown frequency-domain waveform.
    """

    cdef COMPLEX16FrequencySeries *htilde # pointer to waveform structure

    cdef phi0                 # Orbital phase at fRef (rad)
    cdef fRef                 # Reference frequency (Hz)
    cdef deltaF               # Sampling frequency (Hz)
    cdef m1_SI                # Mass of companion 1 (kg)
    cdef m2_SI                # Mass of companion 2 (kg)
    cdef chi1                 # Aligned-spin parameter of companion 1
    cdef chi2                 # Aligned-spin parameter of companion 2
    cdef f_min                # Starting GW frequency (Hz)
    cdef f_max                # End frequency; 0 defaults to Mf = \ref f_CUT
    cdef distance             # Distance of source (m)
    cdef name, length, f0
    cdef fHz, hptilde, hctilde


    def __init__(self, phi0, fRef, deltaF, m1_SI, m2_SI, chi1, chi2, f_min, f_max, distance, inclination):
        """Constructor
        Arguments:
          phi0                  # Orbital phase at fRef (rad)
          fRef_in               # Reference frequency (Hz)
          deltaF                # Sampling frequency (Hz)
          m1_SI                 # Mass of companion 1 (kg)
          m2_SI                 # Mass of companion 2 (kg)
          chi1                  # Aligned-spin parameter of companion 1
          chi2                  # Aligned-spin parameter of companion 2
          f_min                 # Starting GW frequency (Hz)
          f_max                 # End frequency; 0 defaults to Mf = \ref f_CUT
          distance              # Distance of source (m)
          inclination           # Inclination of source (rad)
        """
        # arguments are checked in the C waveform generator
        self.phi0 = phi0
        self.fRef = fRef
        self.deltaF = deltaF
        self.m1_SI = m1_SI
        self.m2_SI = m2_SI
        self.chi1 = chi1
        self.chi2 = chi2
        self.f_min = f_min
        self.f_max = f_max
        self.distance = distance

        self.htilde = NULL

        ret = IMRPhenomDGenerateFD(
            &self.htilde,
            self.phi0,
            self.fRef,
            self.deltaF,
            self.m1_SI,
            self.m2_SI,
            self.chi1,
            self.chi2,
            self.f_min,
            self.f_max,
            self.distance
        );
        if ret == _FAILURE:
            raise ValueError("Call to IMRPhenomDGenerateFD() failed.")

        # Store waveform properties
        self.name = self.htilde.name
        self.length = self.htilde.length
        self.f0 = self.htilde.f0
        #self.deltaF = self.htilde.deltaF
        self.fHz = self.f0 + np.arange(self.length) * self.deltaF

        # Direct copy of C double complex array to numpy via a MemoryView
        cdef double complex[::1] view = <(double complex)[:self.htilde.length]> self.htilde.data
        hptilde = np.asarray(view) # we multiply with the necessary factor below to make this \tilde h_+(f)

        # Produce both polarizations
        cfac = cos(inclination)
        pfac = 0.5 * (1.0 + cfac**2)
        hctilde = np.zeros_like(hptilde)
        hctilde = - 1j * cfac * hptilde
        hptilde *= pfac
        self.hptilde = hptilde
        self.hctilde = hctilde

    def __dealloc__(self):
        """Destructor
        """
        if self.htilde != NULL:
            DestroyCOMPLEX16FrequencySeries(self.htilde)

    def get_waveform(self):
        return np.copy(self.fHz), np.copy(self.hptilde), np.copy(self.hctilde)

# cdef RealVector* ConvertNumpyArrayToRealVector(arr):
#     cdef int n
#     cdef RealVector* vec
#     cdef double* vecdata
#     n = len(arr)
#     vec = CreateRealVector(n)
#     vecdata = vec.data
#     for i in range(n):
#         vecdata[i] = arr[i]
#     return vec


cdef class IMRPhenomDh22AmpPhase:
    """ Generate IMRPhenomD inspiral-merger-ringdown frequency-domain waveform, in Amp/Phase form for h22.
    """

    cdef AmpPhaseFDWaveform* h22 # pointer to waveform structure
    cdef real_vector* Ctf         # pointer to waveform structure
    cdef double fpeak            # Approximate peak frequency (Hz)
    cdef double tpeak            # tf at peak frequency (s)
    cdef double phipeak          # phase 22 at peak frequency
    cdef double fstart           # starting frequency (Hz)
    cdef double tstart           # tf at starting frequency (s)
    cdef double phistart         # phase 22 at starting frequency
    cdef real_vector* Cfreq      # pointer to input frequencies in C structure

    cdef m1                   # Mass of companion 1 (solar masses)
    cdef m2                   # Mass of companion 2 (solar masses)
    cdef chi1                 # Aligned-spin parameter of companion 1 in [-1, 1]
    cdef chi2                 # Aligned-spin parameter of companion 2 in [-1, 1]
    cdef dist                 # Distance of source (Mpc)
    cdef tref                 # Time at fref_for_tref (s)
    cdef phiref               # Orbital phase at fref_for_phiref (rad)
    cdef fref_for_tref        # ref. frequency (Hz) for tref
    cdef fref_for_phiref      # ref. frequency (Hz) for phiref
    cdef force_phiref_fref    # Flag forcing phiref at fref after adjusting tref
    cdef Deltat               # Time shift (s) applied a posteriori
    cdef f_min, f_max
    cdef length
    cdef freq, amp, phase, tf

    cdef ModGRParams* Cmodgrparams # pointer to struct with mod GR params
    cdef ModGRParams Cmodgrparams_struct

    cdef ExtraParams* Cextraparams # pointer to struct with extra params
    cdef ExtraParams Cextraparams_struct

    cdef public object modes  # List of modes (l,m)
    cdef hlm                  # Dictionary of modes
    modes_PhD = [(2,2)]

    def __init__(self,
                 np.ndarray[np.float_t, ndim=1] freq,
                 m1, m2, chi1, chi2, dist,
                 tref=0., phiref=0., fref_for_tref=0., fref_for_phiref=0.,
                 force_phiref_fref=True, Deltat=None,
                 modes=modes_PhD,
                 extra_params=None, mod_gr_params=None):
        """Constructor
        Args:
          freq              # Frequencies (Hz) on which to evaluate the waveform - numpy 1D array
          m1                # Mass of companion 1 (solar masses)
          m2                # Mass of companion 2 (solar masses)
          chi1              # Aligned-spin parameter of companion 1 in [-1,1]
          chi2              # Aligned-spin parameter of companion 2 in [-1,1]
          dist              # Distance of source (Mpc)
        Keyword args:
          tref                # Time at fref_for_tref (s)
          phiref              # Orbital phase at fref_for_phiref (rad)
          fref_for_tref       # ref. frequency (Hz) for tref
          fref_for_phiref     # ref. frequency (Hz) for phiref
          force_phiref_fref   # Flag forcing phiref at fref after adjusting tref
          Deltat              # Time shift (s) applied a posteriori
          modes               # List of modes (l,m) generated - here only (2,2)
          extra_params        # Dictionary of additional parameters
          |- use_buggy_LAL_tpeak # Reproduce bug in tpeak alignment in LAL
          mod_gr_params       # Dictionary of modified GR parameters
        """
        # arguments are checked in the C waveform generator
        self.freq = freq
        self.fpeak = 0.
        self.tpeak = 0.
        self.phipeak = 0.
        self.fstart = 0.
        self.tstart = 0.
        self.phistart = 0.
        self.m1 = m1
        self.m2 = m2
        self.chi1 = chi1
        self.chi2 = chi2
        self.dist = dist
        self.tref = tref
        self.phiref = phiref
        self.fref_for_tref = fref_for_tref
        self.fref_for_phiref = fref_for_phiref
        self.force_phiref_fref = force_phiref_fref
        if Deltat is None:
            Deltat = 0.
        self.Deltat = Deltat
        self.f_min = freq[0]
        self.f_max = freq[-1]
        self.length = len(freq)
        self.modes = modes

        self.h22 = NULL
        self.Ctf = NULL

        # Check input modes
        if not self.modes==[(2,2)]:
            raise ValueError('Only mode (2,2) is available.')

        # Build C structure holding additional params from input python dict
        # if input is None, C struct will stay NULL and code will use default
        self.Cextraparams = NULL
        if extra_params is not None:
            self.Cextraparams_struct.use_buggy_LAL_tpeak = \
                                  extra_params.get('use_buggy_LAL_tpeak', False)
            self.Cextraparams = &self.Cextraparams_struct

        # Build C structure holding modified GR params from input python dict
        # if input is None, C struct will stay NULL and be ignored in C code
        self.Cmodgrparams = NULL
        if mod_gr_params is not None:
            self.Cmodgrparams_struct.alpha = mod_gr_params['alpha'] #PLACEHOLDER
            self.Cmodgrparams = &self.Cmodgrparams_struct

        # Build a real_vector representation of the input numpy array data
        if not freq.flags['C_CONTIGUOUS']:
            raise ValueError('Input numpy array freq is not C_CONTIGUOUS')
        cdef double* freq_data = <double *> &freq[0]
        self.Cfreq = real_vector_view(freq_data, freq.shape[0])

        cdef int Cforce_phiref_fref = <int> self.force_phiref_fref

        ret = IMRPhenomDGenerateh22FDAmpPhase(
            &self.h22,
            &self.Ctf,
            &self.fpeak,
            &self.tpeak,
            &self.phipeak,
            &self.fstart,
            &self.tstart,
            &self.phistart,
            self.Cfreq,
            self.m1,
            self.m2,
            self.chi1,
            self.chi2,
            self.dist,
            self.tref,
            self.phiref,
            self.fref_for_tref,
            self.fref_for_phiref,
            Cforce_phiref_fref,
            self.Deltat,
            self.Cextraparams,
            self.Cmodgrparams
        );
        if ret != 0:
            raise ValueError("Call to IMRPhenomDGenerateFD() failed.")

        # Direct copy of C double array to numpy via a MemoryView
        cdef double[::1] view_amp = <(double)[:self.h22.length]> self.h22.amp
        cdef double[::1] view_phase = \
                                  <(double)[:self.h22.length]> self.h22.phase
        cdef double[::1] view_tf = <(double)[:self.Ctf.size]> self.Ctf.data
        self.amp = np.asarray(view_amp)
        self.phase = np.asarray(view_phase)
        self.tf = np.asarray(view_tf)

    def __dealloc__(self):
        """Destructor
        """
        if self.h22 != NULL:
            DestroyAmpPhaseFDWaveform(self.h22)
        if self.Cfreq != NULL:
            real_vector_view_free(self.Cfreq)
        if self.Ctf != NULL:
            real_vector_free(self.Ctf)

    def get_waveform(self):
        hlm = {}
        lm = (2,2)
        hlm[lm] = {}
        hlm[lm]['freq'] = np.copy(self.freq)
        hlm[lm]['amp'] = np.copy(self.amp)
        hlm[lm]['phase'] = np.copy(self.phase)
        hlm[lm]['tf'] = np.copy(self.tf)
        return hlm

    def get_fpeak(self):
        return self.fpeak

    def get_tpeak(self):
        return self.tpeak

    def get_phipeak(self):
        return self.phipeak

    def get_fstart(self):
        return self.fstart

    def get_tstart(self):
        return self.tstart

    def get_phistart(self):
        return self.phistart

    def compute_toff(self, f):
        cdef int Cforce_phiref_fref = <int> self.force_phiref_fref
        t = IMRPhenomDComputeTimeOfFrequency(
            f,
            self.m1,
            self.m2,
            self.chi1,
            self.chi2,
            self.dist,
            self.tref,
            self.phiref,
            self.fref_for_tref,
            self.fref_for_phiref,
            Cforce_phiref_fref,
            self.Deltat,
            self.Cextraparams,
            self.Cmodgrparams
        );
        return t

    def compute_foft(self, t, f_estimate, t_acc, max_iter=100):
        cdef int Cforce_phiref_fref = <int> self.force_phiref_fref
        cdef int Cmax_iter = <int> max_iter
        f = IMRPhenomDComputeInverseFrequencyOfTime(
            t,
            f_estimate,
            t_acc,
            self.m1,
            self.m2,
            self.chi1,
            self.chi2,
            self.dist,
            self.tref,
            self.phiref,
            self.fref_for_tref,
            self.fref_for_phiref,
            Cforce_phiref_fref,
            self.Deltat,
            Cmax_iter,
            self.Cextraparams,
            self.Cmodgrparams
        );
        return check_error(f)



#
# Copyright (C) 2019 Sylvain Marsat.
#
#


"""
    Cython definitions for spline interpolation tools.
"""

import numpy as np
cimport numpy as np

from lisabeta.struct.pystruct cimport *

cdef extern from "spline.h":

    int BuildNotAKnotSpline(
        real_matrix** splinecoeffs,
        real_vector* vectx,
        real_vector* vecty
    );

    int BuildQuadSpline(
        real_matrix** splinecoeffs,
        real_vector* vectx,
        real_vector* vecty
    );

    int CubicSplineDerivative(
        real_matrix** spline_deriv,
        real_matrix* spline
    );

    int CubicSplineDoubleDerivative(
        real_matrix** spline_dderiv,
        real_matrix* spline
    );

    int CubicSplineIntegral(
        real_matrix** spline_int,
        real_matrix* spline
    );

    int QuadSplineDerivative(
        real_matrix** spline_deriv,
        real_matrix* spline
    );

    int EvalCubicSplineOnVector(
        real_vector** yvec,
        real_vector* xvec,
        real_matrix* spline,
        int extrapol_zero
    );

    int EvalQuadSplineOnVector(
        real_vector** yvec,
        real_vector* xvec,
        real_matrix* spline,
        int extrapol_zero
    );

    # UNFINISHED
    # REQUIRES DEBUGGING
    # int EvalCubicSplineCExpOnConstDeltaxVector(
    #     complex_vector** cexpvec,
    #     real_vector* xvec,
    #     real_matrix* phase_spline
    # );
